@@ -214,6 +214,88 @@ class PgAclListCommand(_AclListHelper):
     lookup_table = 'Port_Group'
 
 
+class AddressSetAddCommand(cmd.AddCommand):
+    table_name = 'Address_Set'
+
+    def __init__(self, api, name, addresses=None, may_exist=False):
+        super().__init__(api)
+        self.name = name
+        self.addresses = [str(netaddr.IPAddress(address))
+                          for address in addresses or []]
+        self.may_exist = may_exist
+
+    def run_idl(self, txn):
+        address_set = self.api.lookup(self.table_name, self.name, None)
+        if address_set:
+            if self.may_exist:
+                self.result = rowview.RowView(address_set)
+                return
+            raise RuntimeError("Address set %s exists" % self.name)
+
+        address_set = txn.insert(self.api.tables[self.table_name])
+        address_set.name = self.name
+        if self.addresses:
+            address_set.addresses = self.addresses
+        self.result = address_set.uuid
+
+
+class AddressSetDelCommand(cmd.BaseCommand):
+    table_name = 'Address_Set'
+
+    def __init__(self, api, address_set, if_exists=False):
+        super().__init__(api)
+        self.address_set = address_set
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            address_set = self.api.lookup(self.table_name, self.address_set)
+            address_set.delete()
+        except idlutils.RowNotFound as e:
+            if self.if_exists:
+                return
+            msg = "Address set %s does not exist" % self.address_set
+            raise RuntimeError(msg) from e
+
+
+class AddressSetGetCommand(cmd.BaseGetRowCommand):
+    table = 'Address_Set'
+
+
+class AddressSetListCommand(cmd.ReadOnlyCommand):
+    table = 'Address_Set'
+
+    def run_idl(self, txn):
+        self.result = [rowview.RowView(r) for
+                       r in self.api.tables[self.table].rows.values()]
+
+
+class AddressSetUpdateAddressesCommand(cmd.BaseCommand):
+    table_name = 'Address_Set'
+
+    def __init__(self, api, address_set, addresses):
+        super().__init__(api)
+        self.address_set = address_set
+        if isinstance(addresses, (str, bytes)):
+            addresses = [addresses]
+        self.addresses = [str(netaddr.IPAddress(address))
+                          for address in addresses]
+
+
+class AddressSetAddAddressesCommand(AddressSetUpdateAddressesCommand):
+    def run_idl(self, txn):
+        address_set = self.api.lookup(self.table_name, self.address_set)
+        for address in self.addresses:
+            address_set.addvalue('addresses', address)
+
+
+class AddressSetRemoveAddressCommand(AddressSetUpdateAddressesCommand):
+    def run_idl(self, txn):
+        address_set = self.api.lookup(self.table_name, self.address_set)
+        for address in self.addresses:
+            address_set.delvalue('addresses', address)
+
+
 class QoSAddCommand(cmd.AddCommand):
     table_name = 'QoS'
 
@@ -815,6 +897,10 @@ class LrpListCommand(cmd.ReadOnlyCommand):
         self.result = [rowview.RowView(r) for r in router.ports]
 
 
+class LrpGetCommand(cmd.BaseGetRowCommand):
+    table = 'Logical_Router_Port'
+
+
 class LrpSetEnabledCommand(cmd.BaseCommand):
     def __init__(self, api, port, is_enabled):
         super().__init__(api)
@@ -845,11 +931,59 @@ class LrpGetOptionsCommand(LspGetOptionsCommand):
     table = 'Logical_Router_Port'
 
 
+class LrpSetGatewayChassisCommand(cmd.BaseCommand):
+    table = 'Logical_Router_Port'
+
+    def __init__(self, api, port, gateway_chassis, priority):
+        super().__init__(api)
+        self.port = port
+        self.gateway_chassis = gateway_chassis
+        self.priority = priority
+
+    def run_idl(self, txn):
+        lrp = self.api.lookup(self.table, self.port)
+        gwc_name = '%s_%s' % (lrp.name, self.gateway_chassis)
+        cmd = GatewayChassisAddCommand(self.api, gwc_name,
+                                       self.gateway_chassis,
+                                       self.priority, may_exist=True)
+        cmd.run_idl(txn)
+        lrp.addvalue('gateway_chassis', cmd.result)
+
+
+class LrpGetGatewayChassisCommand(LrpGetOptionsCommand):
+    def run_idl(self, txn):
+        lrp = self.api.lookup(self.table, self.port)
+        self.result = [rowview.RowView(d) for d in lrp.gateway_chassis]
+
+
+class LrpDelGatewayChassisCommand(cmd.BaseCommand):
+    table = 'Logical_Router_Port'
+
+    def __init__(self, api, port, gateway_chassis, if_exists=False):
+        super().__init__(api)
+        self.port = port
+        self.gateway_chassis = gateway_chassis
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        lrp = self.api.lookup(self.table, self.port)
+        for gwc in lrp.gateway_chassis:
+            if gwc.chassis_name == self.gateway_chassis:
+                lrp.delvalue('gateway_chassis', gwc)
+                return
+
+        if not self.if_exists:
+            msg = "Gateway chassis '%s' on port %s does not exist" % (
+                self.gateway_chassis, self.port)
+            raise RuntimeError(msg)
+
+
 class LrRouteAddCommand(cmd.BaseCommand):
     def __init__(self, api, router, prefix, nexthop, port=None,
                  policy='dst-ip', may_exist=False):
         prefix = str(netaddr.IPNetwork(prefix))
-        nexthop = str(netaddr.IPAddress(nexthop))
+        if nexthop != const.ROUTE_DISCARD:
+            nexthop = str(netaddr.IPAddress(nexthop))
         super().__init__(api)
         self.router = router
         self.prefix = prefix
