@@ -15,7 +15,6 @@ import testscenarios
 
 from ovsdbapp.backend.ovs_idl import idlutils
 from ovsdbapp import constants as const
-from ovsdbapp.schema.ovn_northbound import impl_idl
 from ovsdbapp.tests.functional import base
 from ovsdbapp.tests.functional.schema.ovn_northbound import fixtures
 from ovsdbapp.tests import utils
@@ -27,7 +26,7 @@ class OvnNorthboundTest(base.FunctionalTestCase):
 
     def setUp(self):
         super(OvnNorthboundTest, self).setUp()
-        self.api = impl_idl.OvnNbApiIdlImpl(self.connection)
+        self.api = self.useFixture(fixtures.NbApiFixture(self.connection)).obj
 
 
 class TestLogicalSwitchOps(OvnNorthboundTest):
@@ -236,6 +235,116 @@ class TestAclOps(OvnNorthboundTest):
         self.assertIn(r2, acls)
 
 
+class TestAddressSetOps(OvnNorthboundTest):
+    def setUp(self):
+        super(TestAddressSetOps, self).setUp()
+        self.table = self.api.tables['Address_Set']
+
+    def _addr_set_add(self, name=None, *args, **kwargs):
+        if name is None:
+            name = utils.get_rand_name()
+        fix = self.useFixture(fixtures.AddressSetFixture(self.api, name,
+                                                         *args, **kwargs))
+        self.assertIn(fix.obj.uuid, self.table.rows)
+        return fix.obj
+
+    def _test_addr_set_get(self, col):
+        addr_set = self._addr_set_add()
+        val = getattr(addr_set, col)
+        found = self.api.address_set_get(val).execute(check_error=True)
+        self.assertEqual(addr_set, found)
+
+    def test_addr_set_get_uuid(self):
+        self._test_addr_set_get('uuid')
+
+    def test_addr_set_get_name(self):
+        self._test_addr_set_get('name')
+
+    def test_addr_set_add_name(self):
+        name = utils.get_rand_device_name()
+        addr_set = self._addr_set_add(name)
+        self.assertEqual(name, addr_set.name)
+
+    def test_addr_set_add_exists(self):
+        name = utils.get_rand_device_name()
+        self._addr_set_add(name)
+        cmd = self.api.address_set_add(name)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+    def test_addr_set_add_may_exist(self):
+        name = utils.get_rand_device_name()
+        addr_set = self._addr_set_add(name)
+        addr_set2 = self.api.address_set_add(
+            name, may_exist=True).execute(check_error=True)
+        self.assertEqual(addr_set, addr_set2)
+
+    def test_addr_set_add_with_addresses(self):
+        addresses = ['192.168.0.1', '192.168.0.2']
+        addr_set = self._addr_set_add(addresses=addresses)
+        self.assertEqual(addresses, addr_set.addresses)
+
+    def test_addr_set_del(self):
+        addr_set = self._addr_set_add()
+        self.api.address_set_del(addr_set.uuid).execute(check_error=True)
+        self.assertNotIn(addr_set.uuid, self.table.rows)
+
+    def test_addr_set_del_by_name(self):
+        name = utils.get_rand_device_name()
+        self._addr_set_add(name)
+        self.api.address_set_del(name).execute(check_error=True)
+
+    def test_addr_set_del_no_exist(self):
+        name = utils.get_rand_device_name()
+        cmd = self.api.address_set_del(name)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+    def test_addr_set_del_if_exists(self):
+        name = utils.get_rand_device_name()
+        self.api.address_set_del(
+            name, if_exists=True).execute(check_error=True)
+
+    def test_addr_set_list(self):
+        addr_sets = {self._addr_set_add() for _ in range(3)}
+        found_sets = set(self.api.address_set_list().execute(check_error=True))
+        self.assertTrue(addr_sets.issubset(found_sets))
+
+    def test_addr_set_add_addresses(self):
+        addresses = ['192.168.0.1', '192.168.0.2']
+        addr_set = self._addr_set_add()
+
+        self.api.address_set_add_addresses(
+            addr_set.uuid, addresses).execute(check_error=True)
+        self.assertEqual(addresses, addr_set.addresses)
+
+        self.api.address_set_add_addresses(
+            addr_set.uuid, addresses).execute(check_error=True)
+        self.assertEqual(addresses, addr_set.addresses)
+
+    def test_addr_set_remove_addresses(self):
+        addresses = ['192.168.0.1', '192.168.0.2']
+        addr_set = self._addr_set_add(addresses=addresses)
+
+        self.api.address_set_remove_addresses(
+            addr_set.uuid, addresses).execute(check_error=True)
+        self.assertEqual(addr_set.addresses, [])
+
+        self.api.address_set_remove_addresses(
+            addr_set.uuid, addresses).execute(check_error=True)
+        self.assertEqual(addr_set.addresses, [])
+
+    def test_addr_set_add_remove_addresses_by_str(self):
+        address = "192.168.0.1"
+        addr_set = self._addr_set_add()
+
+        self.api.address_set_add_addresses(
+            addr_set.uuid, address).execute(check_error=True)
+        self.assertEqual([address], addr_set.addresses)
+
+        self.api.address_set_remove_addresses(
+            addr_set.uuid, address).execute(check_error=True)
+        self.assertEqual([], addr_set.addresses)
+
+
 class TestQoSOps(OvnNorthboundTest):
     def setUp(self):
         super(TestQoSOps, self).setUp()
@@ -770,6 +879,11 @@ class TestLogicalRouterOps(OvnNorthboundTest):
         self._lr_add_route(router_name)
         self._lr_add_route(router_name, may_exist=True)
 
+    def test_lr_route_add_discard(self):
+        self._lr_add_route(nexthop=const.ROUTE_DISCARD)
+        self.assertRaises(netaddr.AddrFormatError, self._lr_add_route,
+                          prefix='not-discard')
+
     def test_lr_route_del(self):
         prefix = "192.0.2.0/25"
         route = self._lr_add_route(prefix=prefix)
@@ -1136,6 +1250,18 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
         self.assertEqual(set(networks), set(lrp.networks))
         return lrp
 
+    def _test_lrp_get(self, col):
+        lrp = self._lrp_add(None)
+        val = getattr(lrp, col)
+        found = self.api.lrp_get(val).execute(check_error=True)
+        self.assertEqual(lrp, found)
+
+    def test_lrp_get_uuid(self):
+        self._test_lrp_get('uuid')
+
+    def test_lrp_get_name(self):
+        self._test_lrp_get('name')
+
     def test_lrp_add(self):
         self._lrp_add(None, 'de:ad:be:ef:4d:ad', ['192.0.2.0/24'])
 
@@ -1292,6 +1418,55 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
         self.assertEqual(options, self.api.lrp_get_options(lrp.uuid).execute(
             check_error=True))
 
+    def test_lrp_get_set_gateway_chassis(self):
+        lrp = self._lrp_add(None)
+        c1_name = utils.get_rand_device_name()
+        self.api.lrp_set_gateway_chassis(lrp.uuid, c1_name).execute(
+            check_error=True)
+        c1 = self.api.lookup('Gateway_Chassis', "%s_%s" % (lrp.name, c1_name))
+        lrp_gwcs = self.api.lrp_get_gateway_chassis(lrp.uuid).execute(
+            check_error=True)
+        self.assertIn(c1, lrp_gwcs)
+        self.assertEqual(c1.name, "%s_%s" % (lrp.name, c1_name))
+        self.assertEqual(c1.chassis_name, c1_name)
+        self.assertEqual(c1.priority, 0)
+
+    def test_lrp_set_multiple_gwc(self):
+        lrp = self._lrp_add(None)
+        c1_name, c2_name = [utils.get_rand_device_name() for _ in range(2)]
+        for gwc in [c1_name, c2_name]:
+            self.api.lrp_set_gateway_chassis(lrp.uuid, gwc).execute(
+                check_error=True)
+        c1 = self.api.lookup('Gateway_Chassis', "%s_%s" % (lrp.name, c1_name))
+        c2 = self.api.lookup('Gateway_Chassis', "%s_%s" % (lrp.name, c2_name))
+        lrp_gwcs = self.api.lrp_get_gateway_chassis(lrp.uuid).execute(
+            check_error=True)
+        self.assertIn(c1, lrp_gwcs)
+        self.assertIn(c2, lrp_gwcs)
+
+    def test_lrp_del_gateway_chassis(self):
+        c1_name = utils.get_rand_device_name()
+        lrp = self._lrp_add(None, gateway_chassis=[c1_name])
+        c1 = self.api.lookup('Gateway_Chassis', "%s_%s" % (lrp.name, c1_name))
+        self.assertEqual(lrp.gateway_chassis, [c1])
+        self.api.lrp_del_gateway_chassis(lrp.uuid, c1_name).execute(
+            check_error=True)
+        self.assertEqual(lrp.gateway_chassis, [])
+        self.assertRaises(idlutils.RowNotFound,
+                          self.api.lookup,
+                          'Gateway_Chassis', "%s_%s" % (lrp.name, c1_name))
+
+    def test_lrp_del_gateway_chassis_wrong_chassis(self):
+        lrp = self._lrp_add(None)
+        cmd = self.api.lrp_del_gateway_chassis(lrp.uuid, "fake_chassis")
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+    def test_lrp_del_gateway_chassis_if_exist(self):
+        lrp = self._lrp_add(None)
+        self.api.lrp_del_gateway_chassis(
+            lrp.uuid, "fake_chassis", if_exists=True
+        ).execute(check_error=True)
+
 
 class TestLoadBalancerOps(OvnNorthboundTest):
 
